// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Module: ProductAtts
// Author: Richard Saxton
//
// Description:
//
// A class to provide access to keywords from a SAS input file
// and provide sensible defaults if not available.
//
// Modifications: See ChangeLog
//
// $Id: ProductAtts.cc,v 1.7 2012/03/23 09:25:00 rsaxton Exp $

/* General */

/* SAS-specific */
#include "Tracer.h"
#include "CalServer.h"
#include "Dal.h"
#include "OalServer.h"
#include "ExceptionErrorHandler.h"
#include "ArrayPixelIterator.h"

/* Local */
#include "ProductAtts.h"

/* Constructor */
ProductAtts::ProductAtts(const string & datafile):
_datafile(datafile),
_set(0)
{
  PROC;
      
    /* Open the set */
      _set = dataSetServer->open( datafile, DataSet::Read );

    /* Read the common keywords */

     /* Find the product type and read releveant keywords */
      _productType = "NotSpectrumOrImage";
      if (_set->hasBlock("SPECTRUM")) {
         _productType = "Spectrum";
         //getSpectrumAttributes();
      } else { // Assume that it is an image
          /*
          Array *imgarray=0;
          if (datafile.find(":") != string::npos) {
             imgarray = _set->array(datafile);
          } else {
             imgarray = _set->array(datafile+":");
          }
          if (imgarray->hasAttribute("HDUCLAS1")) {
             if (uppercase(imgarray->attribute("HDUCLAS1")->asString())
                    == "IMAGE") {
                 _productType = "Image";
           // getImageAttributes();
             }
          }
          */
          _productType = "Image";
      }
      
     /* Dates */
      _dateobs = _set->attribute("DATE-OBS")->asString();

      if (_set->hasAttribute("DATE-END"))
      {
         _dateend = _set->attribute("DATE-END")->asString();
      }
      else {
        // No end value so set it to be 10 ks after observation start
         STime numdate(_dateobs);
         numdate = numdate + 10000.0;
         _dateend = numdate.toString();
      }

     /* Mean discarded line value */
      if (_set->hasAttribute("DLMEAN"))
         _dlmean =  _set->attribute("DLMEAN")->asReal();

     /* Exposure time - could be here or could be in spectrum/IMAGE extension */
      if (_set->hasAttribute("EXPOSURE"))
         _exptime =  _set->attribute("EXPOSURE")->asReal();

     /* Selection expression - could be here or could be in spectrum */
      if (_set->hasAttribute("SLCTEXPR"))
         _expression =  _set->attribute("SLCTEXPR")->asString();

     /* Pointing */
      _rapnt  = _set->attribute("RA_PNT")->asReal();
      _decpnt  = _set->attribute("DEC_PNT")->asReal();
      if (_set->hasAttribute("PA_PNT")) {
         _papnt  = _set->attribute("PA_PNT")->asReal();
      } else if (_set->hasAttribute("ROLL_PNT")) {
         _papnt = _set->attribute("ROLL_PNT")->asReal(); // * - 1????
      }

     /* Observation */
      _telescope = _set->attribute("TELESCOP")->asString();
      _datamode = _set->attribute("DATAMODE")->asString();

    /* Get instrument ID */
      _instrument = _set->attribute("INSTRUME")->asString();

      errstr << "Got instrument "
         << _instrument
         << " which has ID "
         << instrumentID()
         << Sas::message(Msg::App, Msg::Noisy);

     /* Get the observing mode and the EMD params */
      readWindowMode();

     /* Set the instrument name in the CAL */
      calServer.setInstrument(instrumentID());

     /* Set the window mode in the cal */
      calServer.state()->mode()->set(_windowMode);

     /* Set the observation date */
      calServer.state()->exposureStartTime()->set(_dateobs);

      if (_set->hasAttribute("EMDHLOW"))
      {
        _emdhlow= _set->attribute("EMDHLOW")->asInt();
      } else {
        _emdhlow=0;
      }

      if (_set->hasAttribute("EMDHUPP"))
      {
        _emdhupp= _set->attribute("EMDHUPP")->asInt();
      } else {
        _emdhupp=4095;
      }

     /* Filter */
      _filter = _set->attribute("FILTER")->asString();

     /* s/w version */
      if (_set->hasAttribute("PPSVERS"))
      {
        _ppsvers = _set->attribute("PPSVERS")->asString();
      } else {
        _ppsvers = "NotSet";
      }

     /* Get the coordinate type if available either in datafile or array header */
      _coordtype = "NotSet"; 
      if (_set->hasAttribute("CTYPE1"))
      {
        _coordtype = _set->attribute("CTYPE1")->asString();
      } else { // Try opening as an array and looking in this header
        if (datafile.find(":") != string::npos)
        {
           // Make sure that this is an array!
           if (_set->block(datafile)->type() == Block::ArrayT) {
               Array *delArr = _set->array(datafile);

               if (delArr->hasAttribute("CTYPE1")) {
                 _coordtype = delArr->attribute("CTYPE1")->asString();
               }  
           } 

        } 
      }

     /* Get any attributes from extensions */
      if ( _productType == "Spectrum")
         getSpectrumAttributes();
      else if ( _productType == "Image")
         getImageAttributes();

     /* Close the file */
      dataSetServer->close(_set);
      //delete _set; _set=0;

  ENDPROC;
}

/* Destructor */
ProductAtts::~ProductAtts()
{
  PROC;
  ENDPROC;
}

/* Method: getSpectrumAttributes()
         : Gets attributes from SPECTRUM header if available */
void ProductAtts::getSpectrumAttributes()
{
  PROC;

   /* Open the spectrum table if it exists */
    Table * spectrumtable;
    if ( _set->hasBlock("SPECTRUM") )
    {
      spectrumtable = _set->table("SPECTRUM");
      
     /* Get selection expression */
      if (spectrumtable->hasAttribute("SLCTEXPR"))
      {
        _expression = spectrumtable->attribute("SLCTEXPR")->asString();
      }
      else
      {
        errstr
        << "Could not find an attribute SLCTEXPR in table "
        << spectrumtable->qualifiedName()
        << ", so assuming SLCEXPR to be true "
        << Sas::message(Msg::App, Msg::Noisy);

        _expression = "true";
      }
    }

    _binsize = spectrumtable->attribute("SPECDELT")->asInt();
    int specpix = spectrumtable->attribute("SPECPIX")->asInt();
    double specval = spectrumtable->attribute("SPECVAL")->asReal();
    int detchans = spectrumtable->attribute("DETCHANS")->asInt();
    _specmin = specval - floor(_binsize / 2.0) - (specpix*_binsize);
    _specmax = _specmin + (detchans * _binsize) -1;

    // Get other keywords
    if (spectrumtable->hasAttribute("EXPOSURE"))
       _exptime =  spectrumtable->attribute("EXPOSURE")->asReal();
    else if (!_exptime)
       errstr<<"Mandatory keyword EXPOSURE missing "
             <<Sas::error("NoExposureTime");

    if (spectrumtable->hasAttribute("DLMEAN"))
       _dlmean =  spectrumtable->attribute("DLMEAN")->asReal();
    else if (!_dlmean)
       errstr<<"Mandatory keyword DLMEAN missing "
             <<Sas::error("NoDLMEAN");

    _backscale=1.0;
    if (spectrumtable->hasAttribute("BACKSCAL"))
       _backscale = spectrumtable->attribute("BACKSCAL")->asReal();

  ENDPROC;
}

int ProductAtts::binsize() const
{
  RETURN _binsize;
}

int ProductAtts::specmin() const
{
  RETURN _specmin;
}

int ProductAtts::specmax() const
{
  RETURN _specmax;
}

/* Method: getImageAttributes()
         : Gets attributes from IMAGE header if available */
void ProductAtts::getImageAttributes()
{
 /* Map the array extension */
    ArrayPixelIterator *_array;
    if (_datafile.find(":") != string::npos) {
       _array = new ArrayPixelIterator(_datafile);
    } else {
       _array = new ArrayPixelIterator(_datafile+":");
    }
    _xbins = _array->xbins();
    _ybins = _array->ybins();

    // Get size of image in physical coords
    _array->first();
    PixelElement px = _array->currentItem();
    _ximagemin = floor(px.xCentre()-px.xWidth()/2.0);
    _yimagemin = floor(px.yCentre()-px.yWidth()/2.0);
    _array->setIndex(_array->size()-1);
    px = _array->currentItem();
    _ximagemax = floor(px.xCentre()+px.xWidth()/2.0);
    _yimagemax = floor(px.yCentre()+px.yWidth()/2.0);

    // Get other keywords
    Array *imgarray=0;
    if (_datafile.find(":") != string::npos) {
       imgarray = _set->array(_datafile);
    } else {
       imgarray = _set->array(_datafile+":");
    }

    if (imgarray->hasAttribute("EXPOSURE"))
       _exptime =  imgarray->attribute("EXPOSURE")->asReal();
    else if (!_exptime)
       errstr<<"Mandatory keyword EXPOSURE missing "
             <<Sas::error("NoExposureTime");

    if (imgarray->hasAttribute("DLMEAN"))
       _dlmean =  imgarray->attribute("DLMEAN")->asReal();
    else if (!_dlmean)
       errstr<<"Mandatory keyword DLMEAN missing "
             <<Sas::error("NoDLMEAN");

    if (imgarray->hasAttribute("SLCTEXPR"))
      {
        _expression = imgarray->attribute("SLCTEXPR")->asString();
      }
      else
      {
       errstr<<"Mandatory keyword SLCTEXPR missing "
             <<Sas::error("NoExpression"); 
      }

    // Get column names
    _xcolumn = imgarray->attribute("CTYPE1L")->asString();
    _ycolumn = imgarray->attribute("CTYPE2L")->asString();

}

int ProductAtts::xbins() const
{
  RETURN _xbins;
}

int ProductAtts::ybins() const
{
  RETURN _ybins;
}

int ProductAtts::ximagemin() const
{
  RETURN _ximagemin;
}
int ProductAtts::ximagemax() const
{
  RETURN _ximagemax;
}
int ProductAtts::yimagemin() const
{
  RETURN _yimagemin;
}
int ProductAtts::yimagemax() const
{
  RETURN _yimagemax;
}

string ProductAtts::xcolumn() const
{
  RETURN _xcolumn;
}
string ProductAtts::ycolumn() const
{
  RETURN _ycolumn;
}
const string ProductAtts::datafile() const
{
  PROC;
  RETURN _datafile;
}

const string   ProductAtts::datamode()   const
{
  PROC;
  RETURN _datamode;
}

const string   ProductAtts::observationDate()   const
{
  PROC;
  RETURN _dateobs;
}

const string   ProductAtts::observationDateEnd()   const
{
  PROC;
  RETURN _dateend;
}

const string   ProductAtts::ppsvers()    const
{
  PROC;
  RETURN _ppsvers;
}

double  ProductAtts::rapnt()  const
{
  PROC;
  RETURN _rapnt;
}

double  ProductAtts::decpnt() const
{
  PROC;
  RETURN _decpnt;
}

double  ProductAtts::papnt()  const
{
  PROC;
  RETURN _papnt;
}

double  ProductAtts::frameTime()  const
{
  PROC;
  RETURN _frameTime;
}

const string   ProductAtts::filter ()    const
{
  PROC;
  RETURN _filter;
}

const string   ProductAtts::instrument()   const
{
  PROC;
  RETURN _instrument;
}

const string   ProductAtts::telescope()   const
{
  PROC;
  RETURN _telescope;
}

const string   ProductAtts::observingMode()   const
{
  PROC;
  RETURN _submode;
}
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * leading const removed
 */
int ProductAtts::emdhlow()   const
{
  PROC;
  RETURN _emdhlow;
}
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * leading const removed
 */
int ProductAtts::emdhupp()   const
{
  PROC;
  RETURN _emdhupp;
}

const string   ProductAtts::coordtype()  const
{
  PROC;
  RETURN _coordtype;
}

/* Method to find the observing window mode */
void ProductAtts::readWindowMode()
{
  PROC;

     /* Read the observing mode from the SUBMODE keyword */
      if (_set->hasAttribute("SUBMODE"))
      {
        _submode = _set->attribute("SUBMODE")->asString();
      } else {
        _submode = " ";
      }

     /* Read the data collection mode from the DATAMODE keyword
      string datamode=" ";
      if (_set->hasAttribute("DATAMODE"))
      {
        datamode = _set->attribute("DATAMODE")->asString();
      }
     */
      errstr<<"Data mode: "<<_datamode
            << Sas::message(Msg::App, Msg::Verbose);

     /* If the observing mode is timing and the datamode is imaging
        then the spectrum has been taken from the outer chips of
        a timing mode obs which are actually in imagind mode */
      if (_datamode=="IMAGING" &&
           (_submode=="FastUncompressed" || _submode=="FastCompressed") )
               { _submode= "PrimeFullWindow"; }

     /* Check if submode string is ok and if not set to default */
      ExceptionErrorHandler local;
      ErrorHandler* oldHandler = errHandler.handler(&local);
      try {
            _windowMode = Xmm::toMode(_submode);
      } catch(SasException::Error) { errstr<<"SUBMODE not valid or missing "<<
          "- using FullWindow" <<Sas::warning("UnknownModeString");
        _submode = "PrimeFullWindow";
        _windowMode = Xmm::toMode(_submode); // d/f
      }
      errHandler.handler(oldHandler);

      errstr<<"Window mode: "<<_windowMode
            << Sas::message(Msg::App, Msg::Verbose);
}

const string ProductAtts::productType() const
{
  RETURN _productType;
}

double ProductAtts::dlmean() const
{
  RETURN _dlmean;
}

double ProductAtts::exptime() const
{
  RETURN _exptime;
}

const string ProductAtts::expression() const
{
  RETURN _expression;
}

Xmm::Instrument ProductAtts::instrumentID()  const
{
  RETURN Xmm::toInstrument(_instrument);
}

double ProductAtts::backscale() const
{
  RETURN _backscale;
}
