// ESA (C) 2000-2018
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Module: QPBselect
// Author: Richard Saxton
//
// Description:
//
// A class to provide access to keywords from a SAS input file
// and provide sensible defaults if not available.
//
// Modifications: See ChangeLog
//
// $Id: QPBselect.cc,v 1.7 2012/03/23 09:25:00 rsaxton Exp $

/* General */

/* SAS-specific */
#include "Tracer.h"
#include "CalServer.h"
#include "Dal.h"
#include "OalServer.h"
#include "ExceptionErrorHandler.h"
#include "metatask.h"
#include "evselect.h"

/* Local */
#include "ProductAtts.h"
#include "QPBselect.h"

//
// This will probably be a C++ metatask but could also be Python
//
using namespace std;
QPBselect::QPBselect()
{
// Read a product, image, spectrum or possibly light curve
// Extract the DLMEAN, DLMAX, exposure time and selection
// expression from the product and set the instrument, filter, 
// observing mode in the CAL
   string prodname = stringParameter("productname");
   string outprod = stringParameter("outprod");
   string input_event_file = stringParameter("table");
   string tempeventfile = stringParameter("outevfile");
   string attfile = stringParameter("attfile");

   // Make the QPB event file using a weight of 1.0
   makeQPBeventfile(input_event_file, tempeventfile, attfile, 1.0);

// Read attributes from input product
   ProductAtts *dfatt = new ProductAtts(prodname);
   
// get the FWCDataServer
   FWCDataServer *fds = 0; fds = calServer.getAtom(fds);

   //vector<double> coeffs = fds->getDLCoeffs(Xmm::PRIME_FULL_WINDOW);

// Get scaling coeffs and put into meaningful variable names
   //unsigned short modeIndex = calServer.instrument()->modeIndex(calServer.state()->mode()->get());
   Xmm::Mode mi = calServer.state()->mode()->get();
   unsigned short modeIndex = calServer.instrument()->modeIndex(mi);
   vector<double> coeffs = fds->getDLCoeffs(modeIndex);
   //cout<<"Mode: "<<modeIndex<<" "<<coeffs[1]<<endl;
   double c0=coeffs[0], c1=coeffs[1], mean_bkg_rate_fwc=coeffs[2],
          exptime_fwc=coeffs[3];

   errstr<<"Scaling coeffs: "<<c0<<" "<<c1<<" "<<mean_bkg_rate_fwc
          <<" "<<exptime_fwc<<Sas::message(Msg::App,Msg::Verbose);

// Calculate scaling factor
   double mean_bkg_rate_sci = (c0 + c1 * dfatt->dlmean());
   double scale_factor = (mean_bkg_rate_sci * dfatt->exptime()) / 
                            (mean_bkg_rate_fwc * exptime_fwc);

   errstr <<"Scale factor: "<<mean_bkg_rate_sci<<" "<<dfatt->exptime()<<" "<<scale_factor<<Sas::message(Msg::App,Msg::Verbose);

   // Close down the FWC calibration atom
   calServer.releaseAtom(fds);
   //DataSet *ds = dataSetServer->clone(input_event_file,"xxx.tmp", Dal::Temp);
// Make a temporary event file 
//   DataSet *ds2 = dataSetServer->clone(input_event_file,"xxx.tmp", Dal::Temp);
   // RDS REINSTATE
   //makeQPBeventfile(input_event_file, tempeventfile, attfile, scale_factor);
 //  dataSetServer->close(ds2);

   // Scale the events`using the EWEIGHT column
   DataSet *ds2 = dataSetServer->open(tempeventfile, Dal::Modify);
   Table *tb = ds2->table("EVENTS");
   ColumnData *cdEW = tb->column("EWEIGHT")->data();
   real64 *eweights = cdEW->real64Data();
    unsigned N = tb->rows();
   for (unsigned n=0; n<N; ++n) 
      eweights[n]=scale_factor;

   delete cdEW;
   dataSetServer->close(ds2);

// Make the QPB product
   DataSet *outset = makeProduct(tempeventfile, outprod, dfatt);

// Scale the counts - no longer needed as scaling factor is 
// written into the temp event file
   //scaleRates(outset, dfatt, scale_factor);

// Update the keywords if necessary
   //updateAttributes(outset, dfatt);

// Close down
   dataSetServer->close(outset);
   delete dfatt; dfatt=0;

} // End of
 
QPBselect::~QPBselect()
{
  // We have shutdown atoms and datasets already
}

// makeQPBeventfile : Method to make a temporary event file of QPB events
void QPBselect::makeQPBeventfile(string & input_event_file, 
                                  string & tempeventfile, 
                                  string & attfile, double scale_factor)
{
  //cout<<input_event_file<<" "<<tempeventfile<<" "<<attfile<<endl;

  // Make the temporary output file
  DataSet *ds = dataSetServer->clone(input_event_file, tempeventfile, Dal::Modify);
  // Check that this is a PN file
  string instr = ds->attribute("INSTRUME")->asString();
  if (Xmm::toInstrument(instr) != Xmm::EPN)
     errstr<<"Wrong instrument: "<<instr
           <<"  - qpbselect only works with EPIC-pn"
           <<Sas::error("WrongInstrument");

  dataSetServer->close(ds);

  Info::Task* evqpb = task("evqpb");
  ParameterInfo& evqpbpars = *evqpb->parameters();
  evqpbpars.revertToDefault();
  evqpbpars["table"]->setValue(input_event_file);
  evqpbpars["outset"]->setValue(tempeventfile);
  evqpbpars["insitu"]->setValue(true);
  evqpbpars["attfile"]->setValue(attfile);
  evqpbpars["overwritesubmode"]->setValue(false);

  // Write the scale factor into the EWEIGHT column
  evqpbpars["withscalefactor"]->setValue(true);
  evqpbpars["scalefactor"]->setValue(scale_factor);

  errstr<<"Running evqpb "<<Sas::message(Msg::App,Msg::Verbose);

  // Run evselect and create the output product
  evqpb->exec();
}

// makeProduct: Method to make the output product
DataSet * QPBselect::makeProduct(string & tempeventfile, 
                                   string & outprod, ProductAtts *dfatt)
{
  Info::Task* evselect = task("evselect");
  ParameterInfo& evselectpars = *evselect->parameters();
  evselectpars.revertToDefault();
  evselectpars["table"]->setValue(tempeventfile);
  evselectpars["expression"]->setValue(dfatt->expression());
  evselectpars["withzcolumn"]->setValue(true);
  evselectpars["zcolumn"]->setValue("EWEIGHT");

  // Set specific parameters depending on the required product
  // Note that the result will a count rate and not counts because
  // scaling will leave floating numbers
  if (dfatt->productType()=="Spectrum") {
     evselectpars["withspectrumset"]->setValue(true);
     evselectpars["spectrumset"]->setValue(outprod);
     evselectpars["energycolumn"]->setValue("PI");
     evselectpars["withspecranges"]->setValue(true);
     evselectpars["spectralbinsize"]->setValue(dfatt->binsize());
     evselectpars["specchannelmin"]->setValue(
                                       float(dfatt->specmin()));
     evselectpars["specchannelmax"]->setValue(
                                       float(dfatt->specmax()));
  } else if (dfatt->productType()=="Image") {
     evselectpars["withimageset"]->setValue(true);
     evselectpars["imageset"]->setValue(outprod);
     //evselectpars["ximagebinsize"]->setValue(dfatt->ximagebinsize());
     //evselectpars["yimagebinsize"]->setValue(dfatt->yimagebinsize());
     evselectpars["ximagesize"]->setValue(dfatt->xbins());
     evselectpars["yimagesize"]->setValue(dfatt->ybins());
     evselectpars["ximagemin"]->setValue(float(dfatt->ximagemin()));
     evselectpars["ximagemax"]->setValue(float(dfatt->ximagemax()));
     evselectpars["yimagemin"]->setValue(float(dfatt->yimagemin()));
     evselectpars["yimagemax"]->setValue(float(dfatt->yimagemax()));
     evselectpars["xcolumn"]->setValue(dfatt->xcolumn());
     evselectpars["ycolumn"]->setValue(dfatt->ycolumn());
  } else {
     errstr<<"Input product of type "<<dfatt->productType()
           <<" can not be processed"<<Sas::error("CantMakeQPBProduct");
  }
  // Run evselect and create the output product
  evselect->exec();

  DataSet * outset = dataSetServer->open(outprod, DataSet::Modify);

  return outset;
}

// Copy header keywords from input to output file (carefully)
void QPBselect::updateAttributes( DataSet * inset,  DataSet * outset)
{
// Attributes in the temporary event set should already be correct
   //copyAttributes(inset, outset);

}
// Scale the rate
void QPBselect::scaleRates(DataSet *outset, ProductAtts *dfatt,
                                                double scale_factor) 
{
   real64     * data; // Data array
   int N=0; // number of elements in spectrum or image
 
   if (dfatt->productType()=="Spectrum") {
      Table * table = outset->table("SPECTRUM");

      Column * dataCol = table->column("RATE");

      ColumnData * dcd = dataCol->data();
      data      = dcd->real64Data();
      N = table->rows();
   } else if (dfatt->productType()=="Image") {
      // Read data from image
   }
   // Scale the data
   for (int n=0; n<N; ++n) 
       data[n] = data[n] * scale_factor;

   // Close down 
}

