// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*
 evqpb

   Authors:  Aitor Ibarra - XMM-Newton SOC
         
   Description: Event list quiescence particle background creator
 
   Comments: 
 
  $Id: module.cc,v 1.5 2000/10/02 08:30:13 gvacanti Exp $
*/

#include "evqpb.h"

#include "stdio.h"
#include <algorithm>
#include <Tracer.h>
#include <errstr.h>
#include <param.h>
#include <Dal.h>
#include <CalServer.h>
#include <STime.h>
#include <Xmm.h>
#include "FWCDataServer.h"
#include "DataFilter.h"
#include "FrameTable.h"
#include "FrameList.h"
#include "metatask.h"
#include "attcalc.h"

using namespace std;

bool compRowData( FWCevent & x,  FWCevent & y)
{
	return (x.time() < y.time());

}


void createTimeStamps( vector<FWCevent> fev, DataSet *ds, int ccds, string instrument, map<unsigned int,double> weights)
{

	 Table *tb = ds->table("EVENTS");

	 /*
	  * Check if we have to add or delete rows in the table.
	  */
	 unsigned long oriSize = tb->rows();
	 unsigned long qpbSize = fev.size();
	 if (tb->rows() > fev.size())
	 {
		 errstr << "Deleting " <<  oriSize- qpbSize <<
				 " rows "<< Sas::message(Msg::App,Msg::Sparse);

		 tb->deleteRows(0,(oriSize- qpbSize));
	 }
	 if (tb->rows() < fev.size())
	 {
		 errstr << "Adding " <<  qpbSize -oriSize  << " rows " <<
				 Sas::message(Msg::App,Msg::Sparse);
		 tb->insertRows(0,(qpbSize -oriSize));
	 }
	 unsigned long newSize = tb->rows();
	 errstr << "New table size " << newSize << Sas::message(Msg::App,Msg::Sparse);

	 //Copy QPB data from CCF to the EVENT extension.
		 ColumnData * timeCdat = tb->column("TIME")->data();
		 real64 * timeDat = timeCdat->real64Data();
		 ColumnData *cdrawx = tb->column("RAWX")->data(),
				 *cdrawy = tb->column("RAWY")->data();
		 int16   *rawX           = cdrawx->int16Data(),
				 *rawY           = cdrawy->int16Data();
	     ColumnData *cddetx = tb->column("DETX")->data(),
	    		 *cddety = tb->column("DETY")->data();
	     int16 *detX = cddetx->int16Data(),
	    		 *detY = cddety->int16Data();
	     ColumnData *cdpi = tb->column("PI")->data();
	     int16 *pi = cdpi->int16Data();
	     ColumnData *cdpha = tb->column("PHA")->data();
	     int16 *pha = cdpha->int16Data();

	     ColumnData *cdflag = tb->column( "FLAG")->data();
	     int32* flag = cdflag->int32Data();
	     ColumnData *cdPattern = tb->column("PATTERN")->data();
	     int8 *pattern = cdPattern->int8Data();
	     ColumnData *cdccdnr = tb->column("CCDNR")->data();
	     int8 *ccdnr = cdccdnr->int8Data();

		 for(unsigned long irow=0; irow< fev.size(); irow++){
			 timeDat[irow] = 0.0;
			 pha[irow] = 0.0;
			 rawX[irow] = fev[irow].rawx();
			 rawY[irow] = fev[irow].rawy();
			 detX[irow] = fev[irow].detx();
			 detY[irow] = fev[irow].dety();
			 pi[irow]	= fev[irow].pi();
			 flag[irow] = fev[irow].flag();
			 pattern[irow] = fev[irow].pattern();
			 ccdnr[irow] = fev[irow].ccd();
		 }


	 // Select events from a ccd
	 for ( int ccdNr = 1 ;  ccdNr <= ccds ; ++ccdNr)
	 {
		 stringstream exposuExt;
		 exposuExt << "EXPOSU";
		 exposuExt.width(2);
		 exposuExt.fill('0');
		 exposuExt << ccdNr;
		 stringstream expression;
		 expression << "CCDNR == " << ccdNr;
		 DataFilter filter(expression.str());

	     vector<DataFilter::RowRange> rowRangeVec;
		 filter.doSelect(*tb,rowRangeVec); // non-destructive filtering

		 if (ds->hasBlock(exposuExt.str()))
		 {
			 Table *exposuTb = ds->table(exposuExt.str());
			 FrameTable in2(exposuTb);
			 FrameList li2; in2>>li2;
			 errstr << "Creating timestamp for ccd " << ccdNr << Sas::message(Msg::App,Msg::Sparse);

			 for (unsigned int j = 0 ; j <= rowRangeVec.size()-1 ; ++j)
			 {
				 for (unsigned int i = rowRangeVec[j].startRow ; i <= rowRangeVec[j].stopRow ; ++i)
				 {
					 fev[i].setTime(li2.getRandomTime());
				 }
			 }
		 }
		 else
		 {
			 errstr << exposuExt.str() << " extension not found " <<
					 Sas::message(Msg::App,Msg::Sparse);
		 }
	 }

	 errstr << "Sorting time column... ";
	 //Short the final time stamps.
	 std::sort(fev.begin(), fev.end(), compRowData);
	 errstr << " Done. " << Sas::message(Msg::App,Msg::Sparse);
	 for(unsigned long irow=0; irow< fev.size(); irow++){
		 timeDat[irow] = fev[irow].time();
	 }

	 ColumnData* eweightDat = tb->addColumn("EWEIGHT", Dal::Real64, "", "Sci/FWC exp weight factor")->data();
	 real64 *eweightData=eweightDat->real64Data();

	 ColumnData* revnoDat = tb->addColumn("REVNO", Dal::Int32, "", "Revolution Number")->data();
	 int32 *revnoData=revnoDat->int32Data();

	 for(unsigned long irow=0; irow< fev.size(); irow++){
		 timeDat[irow] = fev[irow].time();
		 rawX[irow] = fev[irow].rawx();
		 rawY[irow] = fev[irow].rawy();
		 detX[irow] = fev[irow].detx();
		 detY[irow] = fev[irow].dety();
		 pi[irow]	= fev[irow].pi();
		 flag[irow] = fev[irow].flag();
		 pattern[irow] = fev[irow].pattern();
		 ccdnr[irow] = fev[irow].ccd();
		 //cout << "CCDs "<< fev[irow].ccd() <<  " WEIGHTs "  << weights.find(fev[irow].ccd())->second << endl;
		 eweightData[irow] = weights.find(fev[irow].ccd())->second;
		 revnoData[irow] = fev[irow].rev();

	 }


	 delete cdrawx;
	 delete cdrawy;
	 delete cddetx;
	 delete cddety;
	 delete cdpha;
	 delete cdpi;
	 delete cdflag;
	 delete cdccdnr;
	 delete cdPattern;
	 delete timeCdat;
	 delete eweightDat;
	 delete revnoDat;

	 if (Xmm::toInstrument(instrument) == Xmm::EPN)
	 {
		 tb->deleteColumn("PAT_ID");
		 tb->deleteColumn("PAT_SEQ");
		 tb->deleteColumn("TIME_RAW");
	 }

}


void runAttcalc(bool useodf, string evName, string attName, bool ctlmax)
{
	errstr << "Running attcalc to fix X,Y coordinates " << Sas::message(Msg::App,Msg::Sparse);
	Info::Task* attcalc = task("attcalc");
	ParameterInfo& attcalcpars = *attcalc->parameters();
	attcalcpars.revertToDefault();
	attcalcpars["eventset"]->setValue(evName);

        // Use the attitude file and ODF if supplied
        if (useodf) {
	   attcalcpars["withatthkset"]->setValue(true);
	   attcalcpars["atthkset"]->setValue(attName);

        // Get the attitude params from the event file
        } else {

           // Open the output file to get the keywords
           // If any fail throw an error
           DataSet *ds = dataSetServer->open(evName, Dal::Modify);
           try {
             attcalcpars["attitudelabel"]->setValue("fixed");
             attcalcpars["fixedra"]->setValue(ds->attribute("RA_PNT")->asReal());
             attcalcpars["fixeddec"]->setValue(ds->attribute("DEC_PNT")->asReal());
             attcalcpars["fixedposangle"]->setValue(ds->attribute("PA_PNT")->asReal());
             attcalcpars["refpointlabel"]->setValue("user");
             attcalcpars["nominalra"]->setValue(ds->attribute("REFXCRVL")->asReal());
             attcalcpars["nominaldec"]->setValue(ds->attribute("REFYCRVL")->asReal());
             attcalcpars["withatthkset"]->setValue(false);
           } catch(SasException::Error) { errstr<<"Could not get all the "
             <<"attitude info from event file. Please supply an attitude "
             <<"file and ODF"<<Sas::error("AttKeysMissing");
           }
        }

	attcalcpars["calctlmax"]->setValue(ctlmax);	
	attcalc->exec();
}



void evqpb ()
{
PROC

  /*
   * Clone inpupt event file. It will be used as output template
   */
	// Input event file
	string eventFile = stringParameter("table");

	// Exposure factor parameter
	double exposureFactor =  realParameter("exposurefactor");

	// Open DataSet
	string outSet = stringParameter("outset");

        DataSet *ds;
        if (booleanParameter("insitu"))
           ds = dataSetServer->open(outSet, Dal::Modify);
        else
	   ds = dataSetServer->clone(eventFile,outSet, Dal::Modify);

	// Get instrument
	string instr = ds->attribute("INSTRUME")->asString();
	// Set instrument
	//const Xmm::Instrument instrument = Xmm::EPN;
	//calServer.setInstrument(instrument);
	//calServer.state()->ccd()->set(1);


        /* FF, eFF, LW modes now supported (SPR-7564) */
	if (booleanParameter("overwritesubmode") == false)
	{
		Mode epicMode;
		if (ds->hasAttribute("SUBMODE"))
		{
			epicMode = Xmm::toMode(ds->attribute("SUBMODE")->asString());
			switch (epicMode) {
			 case Xmm::PRIME_FULL_WINDOW:
			 //case Xmm::PRIME_FULL_WINDOW_EXTENDED:
			 //case Xmm::PRIME_LARGE_WINDOW:
				 errstr << "Exposure submode: "<< Xmm::toString(epicMode) << Sas::message(Msg::App,Msg::Sparse);
				 break;
			 default:
				 errstr << "Instrument Mode not supported " << Sas::error("WrongInstSubmode");
				 break;
			}
		}

	}
	else
	{
		errstr << "Full frame mode checking disabled. This task should only be used on full frame images"
				<< Sas::warning("SubmodeCheckingDisabled");
	}

	calServer.setInstrument(Xmm::toInstrument(instr));

	STime startTime, stopTime;
	int revolut = 0;
	if (ds->hasAttribute("DATE-OBS"))
		startTime = ds->attribute("DATE-OBS")->asString();
	else
		Sas::error("DATE-OBS keyword not found");

	if (ds->hasAttribute("DATE-END"))
		stopTime = ds->attribute("DATE-END")->asString();
	else
		Sas::error("DATE-END keyword not found");

	if (ds->hasAttribute("REVOLUT"))
		revolut = ds->attribute("REVOLUT")->asInt();
	else
		Sas::error("REVOLUT keyword not found");

	double diff = stopTime - startTime;
	STime tZero = startTime + diff/2.;

  /*
   * Get exposures and noise flags per ccd1 and revno from CAL
   */

	// Get EVENTS table
	Table *tb = ds->table("EVENTS");
	int ccds = calServer.instrument()->nCcds();

	stringstream filterExpr;
	// Accumulative sum of exposure time per CCD

	multimap<int,unsigned int> revNoAndCcds;
	FWCDataServer *fds = 0; fds = calServer.getAtom(fds);
	// Read in the expos time v rev structure
	FWCexptime exps = fds->exptime();
	unsigned int nrevs = exps.rev.size();

	errstr  << "Input Revolution "<< revolut << " startTime " << startTime <<
			"stopTime " << stopTime <<
			Sas::message(Msg::App,Msg::Sparse);


	//Create a sorted mmap with the FWC data middle points
	 multimap<double,unsigned int> fwcMidPoints;
	for (unsigned int i = 0; i <= nrevs -1 ; ++i)
	{
		double diff = (exps.tstart[i]+exps.tstop[i]) /2.0;
		STime revnoTick = STime(Xmm::missionReferenceTime + diff);
		double val = abs(tZero - revnoTick );
		unsigned int rev = exps.rev[i];
		fwcMidPoints.insert(std::pair<double,unsigned int>(val,rev));
	}
	std::multimap<double, unsigned int>::iterator it = fwcMidPoints.begin();

//	for (it = fwcMidPoints.begin(); it != fwcMidPoints.end(); ++it)
//	   {
//	     std::cout << (*it).first << "\t::==\t" << (*it).second  << std::endl;
//	   }



	vector< double >  accumExpTime(ccds,0.0);
	vector< double >  targetExpTime(ccds,0.0);

	//Vector of existing CCDs
	vector< int > validCCDs;

	/*
	 * Get the diferent exposure times
	 * per ccds
	 */
	for (int ccdNr = 1 ;  ccdNr <= ccds ; ++ccdNr)
	{
		stringstream livetime;
		livetime << "LIVETI";
		livetime.width(2);
		livetime.fill('0');
		livetime << ccdNr;
		double exp = 0.0;
		if (tb->hasAttribute(livetime.str()))
		{
			exp = tb->attribute(livetime.str())->asReal();
			// Total exposure time requested by user
			exp *= exposureFactor;
			errstr << "Retrieving "<<  livetime.str() << " Exposure time... "
					<< exp << Sas::message(Msg::App,Msg::Sparse);
			validCCDs.push_back(ccdNr);
		}
		else
		{
			errstr <<livetime.str()<< " Not found" << Sas::message(Msg::App,Msg::Sparse);
		}
		targetExpTime[ ccdNr-1] = exp;
	}

	/*
	 * Calculate total exposure time needed from the FWC data and select
	 * revolutions needed
	 */
	// Accumulate exposure time
	std::vector<int>::iterator firstRevIt = exps.rev.begin();
	std::vector<int>::reverse_iterator lastRevIt = exps.rev.rbegin();
	int ccd = 0;

	while  (  (it !=fwcMidPoints.end() ) )
	{

		// Check if we have accumulated enough exposure time per ccd.
		// And create a vector with ccd that need more exposure time
		std::vector<int> ccdId;
		for ( int ccdNr = 0 ;  ccdNr <= ccds-1 ; ++ccdNr)
		{
                  // Use this CCD if all FWC events are required or if
                  // target exposure time has not yet been reached
                  if ( (booleanParameter("withscalefactor")) || 
                        (targetExpTime[ccdNr] != 0.0 && 
                          accumExpTime[ccdNr] <= targetExpTime[ccdNr] ) )
			{
				ccdId.push_back(ccdNr+1);
				//cout << "CCD " << ccdNr+1 << " NEEDS MORE TIME ACCUM "<< accumExpTime[ccdNr] <<
				//		" TARGET "<< targetExpTime[ccdNr] << endl;
			}

		}

		if (ccdId.size() == 0)
			break;
		//errstr << "Accumulating exposures for ccd "<< ccd+1 << " Target Exposure time: "<< targetExpTime[ccd] <<
		//		" current Accumulated exposure time: " << accumExpTime[ccd] << Sas::message(Msg::App,Msg::Sparse);

		// Find index for first for closet

		lastRevIt  = std::find(exps.rev.rbegin(), exps.rev.rend(),(*it).second );
		firstRevIt  = std::find(exps.rev.begin(), exps.rev.end(),(*it).second );

		unsigned int firstElement = std::distance(exps.rev.begin(), firstRevIt);
		unsigned int lastElement = std::distance(exps.rev.rbegin(), lastRevIt);
		if(firstRevIt != exps.rev.end())
		{
			errstr << "REV found  " << exps.rev[firstElement] <<  " index " <<  firstElement <<
					" lastElement "<< nrevs-lastElement<<
					Sas::message(Msg::App,Msg::Sparse);
		}
		else
		{
			errstr << "Revolution not found " << Sas::message(Msg::App,Msg::Always);
		}

		for ( int ccdNr = 1 ;  ccdNr <= ccds ; ++ccdNr)
		{
			//errstr << "Accumulating time per ccd: "<< ccdNr
			//		<< " target exposure " << targetExpTime[ccdNr-1]<<
			//		Sas::message(Msg::App,Msg::Sparse);

			//std::cout << (*it).first << "\t::==\t" << (*it).second  << std::endl;
			// Build multimap with revno and not noisy ccds to pass to CAL
			if ( (ccdNr == exps.ccd[firstElement]) && (exps.nflag[ccdNr] == 0) )
			{
				if ( std::find(ccdId.begin(), ccdId.end(), ccdNr) != ccdId.end() )
				{
					int revNo = exps.rev[firstElement];
					accumExpTime[ccdNr-1] = accumExpTime[ccdNr-1] + exps.exptime[firstElement];

					errstr << "Rev. Number: " << revNo << " ccd " << exps.ccd[firstElement] <<
							" current accum exposure " << accumExpTime[ccdNr-1] <<
							Sas::message(Msg::App,Msg::Sparse);

					revNoAndCcds.insert(pair<int, unsigned int>(revNo,ccdNr));
				}
			}
			++firstElement;
			++it;
		}
		++ccd;
	}




	//MAP of ccds and weights
	map<unsigned int, double> ccdsWeights;

	for ( unsigned int i = 0 ; i < unsigned(ccds) ; ++i )
	{
		stringstream str;
		str << "EWEIGHT";
		str.width(2);
		str.fill('0');
		str << i+1;
		if (accumExpTime[i] != 0.0)
		{
                  // Set the weight for this CCD to the exposure time
                  // fraction or to a fixed value from input parameter
		   string eventFile = stringParameter("table");
                   double val = 0.0;
                   if (booleanParameter("withscalefactor")) {
                      val=realParameter("scalefactor");
                   } else {
		      val = (targetExpTime[i]/exposureFactor )/accumExpTime[i];
		      errstr << "CCD " << i+1 <<
				" EXPOSURE " << targetExpTime[i]/exposureFactor <<
				" TARGET "<<  targetExpTime[i] << " TOTAL EXP " << accumExpTime[i] <<
				" Weight Factor "<< val <<
	           			Sas::message(Msg::App,Msg::Sparse);
                   } 
		   ccdsWeights.insert(std::pair<unsigned int, double>(i+1,val));

			//*(tb->addAttribute(str.str(),"Sci/FWC Exp Weight factor","")) = val;
		}
                //cout<<"Done eweight for CCD: "<<i+1<<endl;
	}

//	 for (auto& x: ccdsWeights) {
//	    std::cout << x.first << ": " << x.second << '\n';
//	  }

/*
	std::multimap<int, unsigned int>::iterator it2;
	 for (it2 = revNoAndCcds.begin(); it2 != revNoAndCcds.end(); ++it2)
	   {
	     std::cout << (*it2).first << "\t::==\t" << (*it2).second  << std::endl;
	   }
*/
	 vector<FWCevent> fev = fds->getEvents(revNoAndCcds);



  /*
   * Create new time stamps for FWC events using EXPOSU extensions
   */
	 createTimeStamps(fev,ds,ccds,instr,ccdsWeights );

  /*
   * run attcalc
   */
	 errstr << "running attcalc" << Sas::message(Msg::App,Msg::Sparse);
         bool useodf = booleanParameter("useodfatt");
	 string attFileName = stringParameter("attfile");
	 bool calctlmax = booleanParameter("calctlmax"); 
	 runAttcalc(useodf,outSet,attFileName,calctlmax);

	 if (ds->hasAttribute("FILTER"))
		 *ds->addAttribute("FILTER"," Filter ID") = Xmm::toString(Xmm::CLOSED);

	 int blks = ds->blocks();

	 for (int i = 0; i < blks; ++i)
		 if (ds->block(i)->hasAttribute("FILTER"))
			 *ds->block(i)->addAttribute("FILTER","Filter ID") = Xmm::toString(Xmm::CLOSED);



  /*
   * Customize keywords and file name.
   */ 
	 dataSetServer->close(ds);

ENDPROC
}
